<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub inbox requests
 */
class Inbox implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main api main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch inbox for @unused-param
     * @return string|false response
     */
    public function get($_user)
    {
        return false;
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to inbox
     * @return string|false response
     */
    public function post($_user)
    {
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            error_log("Inbox::post Signature check failed: " . $e->getMessage());
            http_response_code(401);
            return false;
        }

        $activity = is_string($_rawInput) ? json_decode($_rawInput, true) : null;
        $host = $_SERVER['SERVER_NAME'];

        if (!is_array($activity)) {
            error_log("Inbox::post Input wasn't of type array");
            return false;
        }

        $inboxActivity = \Federator\Data\ActivityPub\Factory::newActivityFromJson($activity);

        // Shared inbox
        if (!isset($_user)) {
            $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';

            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/inbox.log',
                date('Y-m-d H:i:s') . ": ==== POST Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }

        if ($inboxActivity === false) {
            error_log("Inbox::post couldn't create inboxActivity, aborting");
            return false;
        }

        $sendTo = $inboxActivity->getCC();
        if ($inboxActivity->getType() === 'Undo') { // for undo the object holds the proper cc
            $object = $inboxActivity->getObject();
            if ($object !== null) {
                $sendTo = $object->getCC();
            }
        }

        $users = [];

        foreach ($sendTo as $receiver) {
            if ($receiver === '' || !is_string($receiver)) {
                continue;
            }

            if (str_ends_with($receiver, '/followers')) {
                $followers = $this->fetchAllFollowers($receiver, $host);
                if (is_array($followers)) {
                    $users = array_merge($users, $followers);
                }
            }
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }
        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }

            $this->postForUser($user, $inboxActivity);
        }
        return json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call for specific user
     *
     * @param string $_user user to add data to inbox
     * @param \Federator\Data\ActivityPub\Common\Activity $inboxActivity the activity that we received
     * @return boolean response
     */
    private function postForUser($_user, $inboxActivity)
    {
        if (isset($_user)) {
            $dbh = $this->main->getDatabase();
            $cache = $this->main->getCache();
            $connector = $this->main->getConnector();

            // get user
            $user = \Federator\DIO\User::getUserByName(
                $dbh,
                $_user,
                $connector,
                $cache
            );
            if ($user->id === null) {
                error_log("Inbox::postForUser couldn't find user: $_user");
                return false;
            }
        }

        $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/inbox_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        return true;
    }

    /**
     * fetch all followers from url and return the ones that belong to our server
     *
     * @param string $collectionUrl The url of f.e. the posters followers
     * @param string $host our current host-url
     * @return string[] the names of the followers that are hosted on our server
     */
    private static function fetchAllFollowers(string $collectionUrl, string $host): array
    {
        $users = [];

        [$collectionResponse, $collectionInfo] = \Federator\Main::getFromRemote($collectionUrl, ['Accept: application/activity+json']);
        if ($collectionInfo['http_code'] != 200) {
            error_log("Inbox::fetchAllFollowers Failed to fetch follower collection metadata from $collectionUrl");
            return [];
        }

        $collectionData = json_decode($collectionResponse, true);
        $nextPage = $collectionData['first'] ?? $collectionData['current'] ?? null;

        if (!isset($nextPage)) {
            error_log("Inbox::fetchAllFollowers No 'first' or 'current' page in collection at $collectionUrl");
            return [];
        }

        // Loop through all pages
        while ($nextPage) {
            [$pageResponse, $pageInfo] = \Federator\Main::getFromRemote($nextPage, ['Accept: application/activity+json']);
            if ($pageInfo['http_code'] != 200) {
                error_log("Inbox::fetchAllFollowers Failed to fetch follower page at $nextPage");
                break;
            }

            $pageData = json_decode($pageResponse, true);
            $items = $pageData['orderedItems'] ?? $pageData['items'] ?? [];

            foreach ($items as $followerUrl) {
                $parts = parse_url($followerUrl);
                if (!isset($parts['host']) || !str_ends_with($parts['host'], $host)) {
                    continue;
                }

                [$actorResponse, $actorInfo] = \Federator\Main::getFromRemote($followerUrl, ['Accept: application/activity+json']);
                if ($actorInfo['http_code'] != 200) {
                    error_log("Inbox::fetchAllFollowers Failed to fetch actor data for follower: $followerUrl");
                    continue;
                }

                $actorData = json_decode($actorResponse, true);
                if (isset($actorData['preferredUsername'])) {
                    $users[] = $actorData['preferredUsername'];
                }
            }

            $nextPage = $pageData['next'] ?? null;
        }

        return $users;
    }
}
