<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\V1;

/**
 * Called from our application to inform us about new content (f.e. new posts on contentnation.net)
 */
class NewContent implements \Federator\Api\APIInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * response from sub-calls
     *
     * @var string $response
     */
    private $response;

    /**
     * constructor
     *
     * @param \Federator\Main $main main instance
     * @return void
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * run given url path
     *
     * @param array<string> $paths path array split by /
     * @param \Federator\Data\User|false $user user who is calling us @unused-param
     * @return bool true on success
     */
    public function exec($paths, $user)
    {
        $method = $_SERVER["REQUEST_METHOD"];
        $_username = $paths[2];
        if ($method === 'GET') { // unsupported
            throw new \Federator\Exceptions\InvalidArgument("GET not supported");
        }
        switch (sizeof($paths)) {
            case 3:
                $ret = $this->post($_username);
                break;
        }

        if (isset($ret) && $ret !== false) {
            $this->response = $ret;
            return true;
        }

        $this->main->setResponseCode(404);
        return false;
    }


    /**
     * handle post call
     *
     * @param string|null $_user user that triggered the post
     * @return string|false response
     */
    public function post($_user)
    {
        error_log("NewContent::post called with user: $_user");
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            error_log("NewContent::post Signature check failed: " . $e->getMessage());
            http_response_code(401);
            return false;
        }

        $input = is_string($_rawInput) ? json_decode($_rawInput, true) : null;
        $host = $_SERVER['SERVER_NAME'];
        if (!is_array($input)) {
            error_log("NewContent::post Input wasn't of type array");
            return false;
        }

        if (isset($allHeaders['X-Sender'])) {
            $newActivity = $this->main->getConnector()->jsonToActivity($input);
        } else {
            $newActivity = \Federator\Data\ActivityPub\Factory::newActivityFromJson($input);
        }

        if ($newActivity === false) {
            error_log("NewContent::post couldn't create newActivity");
            return false;
        }

        $sendTo = $newActivity->getCC();
        if ($newActivity->getType() === 'Undo') {
            $object = $newActivity->getObject();
            if ($object !== null) {
                $sendTo = $object->getCC();
            }
        }

        $users = [];

        foreach ($sendTo as $receiver) {
            if ($receiver === '' || !is_string($receiver)) {
                continue;
            }

            if (str_ends_with($receiver, '/followers')) {
                $followers = $this->fetchAllFollowers($receiver, $host);
                if (is_array($followers)) {
                    $users = array_merge($users, $followers);
                }
            }
        }

        if (empty($users)) { // todo remove, debugging for now
            $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';
            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/newContent.log',
                date('Y-m-d H:i:s') . ": ==== POST NewContent Activity ====\n" . json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }

        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }

            $this->postForUser($user, $newActivity);
        }

        return json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call for specific user
     *
     * @param string $_user user that triggered the post
     * @param \Federator\Data\ActivityPub\Common\Activity $newActivity the activity that we received
     * @return boolean response
     */
    private function postForUser($_user, $newActivity)
    {
        if (isset($_user)) {
            $dbh = $this->main->getDatabase();
            $cache = $this->main->getCache();
            $connector = $this->main->getConnector();

            // get user
            $user = \Federator\DIO\User::getUserByName(
                $dbh,
                $_user,
                $connector,
                $cache
            );
            if ($user->id === null) {
                error_log("NewContent::postForUser couldn't find user: $_user");
                return false;
            }
        }

        $rootDir = $_SERVER['DOCUMENT_ROOT'] . '../';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/newcontent_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " NewContent Activity ====\n" . json_encode($newActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        return true;
    }

    /**
     * fetch all followers from url and return the ones that belong to our server
     *
     * @param string $collectionUrl The url of f.e. the posters followers
     * @param string $host our current host-url
     * @return string[] the names of the followers that are hosted on our server
     */
    private static function fetchAllFollowers(string $collectionUrl, string $host): array
    {
        $users = [];

        [$collectionResponse, $collectionInfo] = \Federator\Main::getFromRemote($collectionUrl, ['Accept: application/activity+json']);
        if ($collectionInfo['http_code'] != 200) {
            error_log("NewContent::fetchAllFollowers Failed to fetch follower collection metadata from $collectionUrl");
            return [];
        }

        $collectionData = json_decode($collectionResponse, true);
        $nextPage = $collectionData['first'] ?? $collectionData['current'] ?? null;

        if (!isset($nextPage)) {
            error_log("NewContent::fetchAllFollowers No 'first' or 'current' page in collection at $collectionUrl");
            return [];
        }

        // Loop through all pages
        while ($nextPage) {
            [$pageResponse, $pageInfo] = \Federator\Main::getFromRemote($nextPage, ['Accept: application/activity+json']);
            if ($pageInfo['http_code'] != 200) {
                error_log("NewContent::fetchAllFollowers Failed to fetch follower page at $nextPage");
                break;
            }

            $pageData = json_decode($pageResponse, true);
            $items = $pageData['orderedItems'] ?? $pageData['items'] ?? [];

            foreach ($items as $followerUrl) {
                $parts = parse_url($followerUrl);
                if (!isset($parts['host']) || !str_ends_with($parts['host'], $host)) {
                    continue;
                }

                [$actorResponse, $actorInfo] = \Federator\Main::getFromRemote($followerUrl, ['Accept: application/activity+json']);
                if ($actorInfo['http_code'] != 200) {
                    error_log("NewContent::fetchAllFollowers Failed to fetch actor data for follower: $followerUrl");
                    continue;
                }

                $actorData = json_decode($actorResponse, true);
                if (isset($actorData['preferredUsername'])) {
                    $users[] = $actorData['preferredUsername'];
                }
            }

            $nextPage = $pageData['next'] ?? null;
        }

        return $users;
    }

    /**
     * get internal represenation as json string
     * @return string json string or html
     */
    public function toJson()
    {
        return $this->response;
    }
}