<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Sascha Nitsch (grumpydeveloper)
 **/

namespace Federator\Data\ActivityPub\Common;

class APObject implements \JsonSerializable
{
    // actor | bto | current | first | id | instrument | last | items | oneOf |
    // anyOf | closed | origin | next | object | prev | result | target |
    // type | accuracy | altitude | content | duration | height | href | hreflang |
    // partOf | latitude | longitude | endTime | radius | rel |
    // startIndex | totalItems | units | width | subject | relationship | describes |
    // formerType | deleted

    /**
     * unique id
     *
     * @var string $id
     */
    private $id = '';

    /**
     * child object
     *
     * @var APObject|null $object
     */
    private $object = null;

    /**
     * type
     *
     * @var string $type
     */
    private $type = 'Object';

    /**
     * content
     *
     * @var string $content
     */
    private $content = '';

    /**
     * duration in seconds
     *
     * @var \DateInterval|false $duration
     */
    private $duration = false;

    /**
     * height
     *
     * @var int $height
     */
    private $height = -1;

    /**
     * href
     *
     * @var string $href
     */
    private $href = '';

    /**
     * end time
     *
     * @var int $endTime
     */
    private $endTime = 0;

    /**
     * width
     *
     * @var int $width
     */
    private $width = -1;


    // fiels are attachment | attributedTo | audience | content | context | name | endTime | generator |
    // icon | image | inReplyTo | location | preview | published | replies | startTime | summary | tag |
    // updated | url | to | bto | cc | bcc | mediaType | duration

    /**
     * attachements
     *
     * @var APObject[] $attachment
     */
    private $attachment = [];

    /**
     * attributed to
     *
     * @var string $attributedTo
     */
    private $attributedTo = '';

    // audience
    // content
    // context

    /**
     * name
     *
     * @var string $name
     */
    private $name = '';

    // endTime
    // generator

    /**
     * images
     * @var Image[] $icon
     */
    private $icon = array();

    /**
     * images
     *
     * @var Image[] $image
     */
    private $image = array();

    /**
     * reply id
     *
     * @var string $inReplyTo
     */
    private $inReplyTo = "";

    // location
    // preview

    /**
     * published timestamp
     *
     * @var int $published
     */
    private $published = 0;

    // startTime

    /**
     * summary
     *
     * @var string $summary
     */
    private $summary = "";

    /**
     * tags
     *
     * @var Tag[] $tag
     */
    private $tag = array();

    /**
     * updated timestamp
     *
     * @var int $updated
     */
    private $updated = 0;

    /**
     * url
     *
     * @var string $url
     */
    private $url = "";

    /**
     * list of to ids
     *
     * @var array<string> $to
     */
    private $to = array();

    // bto

    /**
     * list of cc ids
     *
     * @var array<string> $cc
     */
    private $cc = array();

    // bcc

    /**
     * media type
     *
     * @var string $mediaType
     */
    private $mediaType = "";

    // duration

    /**
     * list of contexts
     *
     * @var array<string> $context
     */
    private $context = [];

    /**
     * key value map of contexts
     * @var array<string, string> $contexts
     */
    private $contexts = [];

    /**
     * atom URI
     *
     * @var string $atomURI
     */
    private $atomURI = '';

    /**
     * reply to atom URI
     *
     * @var ?string $replyAtomURI
     */
    private $replyAtomURI = null;

    // found items

    /**
     * generated unique id
     *
     * @var string $uuid
     */
    private $uuid = "";

    /**
     * sensitive flag
     *
     * @var ?bool $sensitive
     */
    private $sensitive = null;

    /**
     * blur hash
     * @var string $blurhash
     */
    private $blurhash = "";

    /**
     * conversation id
     * @var string $conversation
     */
    private $conversation = '';

    /**
     * constructor
     *
     * @param string $type type of object
     */
    public function __construct(string $type)
    {
        $this->type = $type;
    }

    /**
     * set note content
     *
     * @param string $content note content
     * @return APObject current instance
     */
    public function setContent(string $content)
    {
        $this->content = $content;
        return $this;
    }

    /**
     * get note content
     *
     * @return string
     */
    public function getContent()
    {
        return $this->content;
    }

    /**
     * add context to list
     *
     * @param string $context new context
     * @return void
     */
    final public function addContext($context)
    {
        if (!in_array($context, $this->context, false)) {
            $this->context[] = $context;
        }
    }

    /**
     * add multiple contexts to list
     *
     * @param array<string,string> $contexts new contexts
     * @return void
     */
    final public function addContexts($contexts)
    {
        foreach ($contexts as $key => $value) {
            if (!in_array($key, $this->contexts, false)) {
                $this->contexts[$key] = $value;
            }
        }
    }

    /**
     * set id
     *
     * @param string $id new id
     * @return APObject current object
     */
    final public function setID($id)
    {
        $this->id = $id;
        return $this;
    }

    final public function getID() : string
    {
        return $this->id;
    }

    /**
     * set href
     *
     * @param string $href href
     * @return APObject
     */
    public function setHref(string $href)
    {
        $this->href = $href;
        return $this;
    }

    /**
     * set child object
     *
     * @param APObject $object
     * @return void
     */
    public function setObject($object)
    {
        $this->object = $object;
    }

    /**
     * get child object
     *
     * @return APObject|null child object
     */
    public function getObject()
    {
        return $this->object;
    }

    /**
     * set summary
     *
     * @param string $summary summary
     * @return APObject this
     */
    public function setSummary($summary)
    {
        $this->summary = $summary;
        return $this;
    }
    /**
     * set type
     *
     * @param string $type type
     * @return void
     */
    public function setType(string $type)
    {
        $this->type = $type;
    }

    public function getType() : string
    {
        return $this->type;
    }

    /**
     * set attachments
     *
     * @param APObject[] $attachment
     * @return void
     */
    public function setAttachment($attachment)
    {
        $this->attachment = $attachment;
    }

    public function addAttachment(APObject $attachment) : void
    {
        $this->attachment[] = $attachment;
    }

    /**
     * get attachments
     *
     * @return APObject[] attachments
     */
    public function getAttachment()
    {
        return $this->attachment;
    }

    public function getAttachmentsAsJson() : string
    {
        if ($this->attachment === []) {
            return "{}";
        }
        $obj = [];
        foreach ($this->attachment as $a) {
            $obj[] = $a->toObject();
        }
        return json_encode($obj) | '';
    }

    /**
     * set attributed to
     *
     * @param string $to attribute to
     * @return APObject
     */
    public function setAttributedTo(string $to)
    {
        $this->attributedTo = $to;
        return $this;
    }

    public function getAttributedTo() : string
    {
        return $this->attributedTo;
    }

    /**
     * set name
     *
     * @param string $name name
     * @return APObject
     */
    public function setName(string $name)
    {
        $this->name = $name;
        return $this;
    }

    /**
     * add Image
     *
     * @param Image $image image to add
     * @return void
     */
    public function addImage(Image $image)
    {
        $this->image[] = $image;
    }

    public function setInReplyTo(string $reply) : void
    {
        $this->inReplyTo = $reply;
    }

    public function getInReplyTo() : string
    {
        return $this->inReplyTo ?? "";
    }

    /**
     * set published timestamp
     *
     * @param int $published published timestamp
     * @return APObject
     */
    public function setPublished(int $published)
    {
        $this->published = $published;
        return $this;
    }

    public function getPublished() : int
    {
        return $this->published;
    }

    /**
     * add Tag
     *
     * @param Tag $tag tag to add
     * @return void
     */
    public function addTag(Tag $tag)
    {
        $this->tag[] = $tag;
    }

    /**
     * set url
     *
     * @param string $url URL
     * @return APObject
     */
    public function setURL(string $url)
    {
        $this->url = $url;
        return $this;
    }

    /**
     * get URL
     *
     * @return string URL
     */
    public function getURL()
    {
        return $this->url;
    }

    /**
     * add to
     *
     * @param string $to additional to address
     * @return APObject
     */
    public function addTo(string $to)
    {
        $this->to[] = $to;
        return $this;
    }

    /**
     * get to
     *
     * @return array<string>
     */
    public function getTo()
    {
        return $this->to;
    }

    /**
     * add cc
     *
     * @param string $cc additional cc address
     * @return void
     */
    public function addCC($cc)
    {
        $this->cc[] = $cc;
    }

    /**
     * get cc
     *
     * @return array<string>
     */
    public function getCC()
    {
        return $this->cc;
    }

    public function getMediaType() : string
    {
        return $this->mediaType;
    }

    /**
     * set atom URI
     *
     * @param string $uri atom URI
     * @return void
     */
    public function setAtomURI($uri)
    {
        $this->atomURI = $uri;
    }

    /**
     * set reply atom URI
     * @param string $uri reply atom URI
     * @return void
     */
    public function setReplyAtomURI($uri)
    {
        $this->replyAtomURI = $uri;
    }

    /**
     * set sensitive
     *
     * @param bool $sensitive status
     * @return void
     */
    public function setSensitive($sensitive)
    {
        $this->sensitive = $sensitive;
    }

    /**
     * set conversation id
     * @param string $conversation conversation ID
     * @return void
     */
    public function setConversation(string $conversation)
    {
        $this->conversation = $conversation;
    }

    public function getConversation() : ?string
    {
        return $this->conversation;
    }

    /**
     * create from json
     *
     * @param array<string, mixed> $json input
     * @return bool true on success
     */
    public function fromJson($json)
    {
        if (!is_array($json)) {
            error_log("fromJson called with ".gettype($json). " => ". debug_backtrace()[1]['function']
             . " json: " . print_r($json, true));
            return false;
        }
        if (array_key_exists('id', $json)) {
            $this->id = $json['id'];
        }
        if (array_key_exists('content', $json)) {
            $this->content = $json['content'];
        }
        if (array_key_exists('duration', $json)) {
            try {
                $this->duration = new \DateInterval($json['duration']);
            } catch (\Exception $unused_e) {
                error_log("error parsing duration ". $json['duration']);
            }
        }
        if (array_key_exists('height', $json)) {
            $this->height = intval($json['height'], 10);
        }
        if (array_key_exists('href', $json)) {
            $this->href = $json['href'];
        }
        if (array_key_exists('endTime', $json) && $json['endTime'] !== null) {
            $this->endTime = $this->parseDateTime($json['endTime']);
        }
        if (array_key_exists('width', $json)) {
            $this->width = intval($json['width'], 10);
        }
        if (array_key_exists('attachment', $json) && $json['attachment'] !== null) {
            $attachment = [];
            foreach ($json['attachment'] as $a) {
                $att = \Federator\Data\ActivityPub\Factory::newFromJson($a, "");
                if ($att !== null) {
                    $attachment[] = $att;
                }
            }
            $this->attachment = $attachment;
        }
        if (array_key_exists('attributedTo', $json)) {
            if (is_array($json['attributedTo']) && array_key_exists(0, $json['attributedTo'])) {
                if (is_array($json['attributedTo'][0])) {
                    $this->attributedTo = $json['attributedTo'][0]['id'];
                } else {
                    $this->attributedTo = $json['attributedTo'][0];
                }
            } else {
                $this->attributedTo = (string)$json['attributedTo'];
            }
        }
        if (array_key_exists('name', $json)) {
            $this->name = $json['name'];
        }
        if (array_key_exists('icon', $json)) {
            if (array_key_exists('type', $json['icon'])) {
                $image = new Image();
                $image->fromJson($json['icon']);
                $this->icon[] = $image;
            } else {
                foreach ($json['icon'] as $icon) {
                    $image = new Image();
                    $image->fromJson($icon);
                    $this->icon[] = $image;
                }
            }
        }
        if (array_key_exists('image', $json)) {
            $image = new Image();
            $image->fromJson($json['image']);
            $this->image[] = $image;
        }
        if (array_key_exists('inReplyTo', $json)) {
            $this->inReplyTo = $json['inReplyTo'];
        }
        if (array_key_exists('published', $json)) {
            $this->published = $this->parseDateTime($json['published']);
        }
        if (array_key_exists('summary', $json)) {
            $this->summary = $json['summary'];
        }
        if (array_key_exists('tag', $json)) {
            $tags = [];
            foreach ($json['tag'] as $t) {
                $tag = new Tag();
                $tag->fromJson($t);
                $tags[] = $tag;
            }
            $this->tag = $tags;
        }
        if (array_key_exists('updated', $json)) {
            $this->updated = $this->parseDateTime($json['updated']);
        }
        if (array_key_exists('url', $json)) {
            $this->url = $json['url'];
        }
        if (array_key_exists('to', $json)) {
            if (is_array($json['to'])) {
                foreach ($json['to'] as $to) {
                    $this->to[] = $to;
                }
            } else {
                $this->to[] = $json['to'];
            }
        }
        if (array_key_exists('cc', $json)) {
            if (is_array($json['cc'])) {
                foreach ($json['cc'] as $cc) {
                    $this->cc[] = $cc;
                }
            } else {
                $this->cc[] = $json['cc'];
            }
        }
        if (array_key_exists('mediaType', $json)) {
            $this->mediaType = $json['mediaType'];
        }
        if (array_key_exists('object', $json)) {
            $this->object = \Federator\Data\ActivityPub\Factory::newFromJson($json['object'], "");
        }
        if (array_key_exists('sensitive', $json)) {
            $this->sensitive = $json['sensitive'];
        }
        if (array_key_exists('blurhash', $json)) {
            $this->blurhash = $json['blurhash'];
        }
        if (array_key_exists('uuid', $json)) {
            $this->uuid = $json['uuid'];
        }
        if (array_key_exists('conversation', $json)) {
            $this->conversation = $json['conversation'];
        }
        return true;
    }

    /**
     *
     * {@inheritDoc}
     * @see JsonSerializable::jsonSerialize()
     */
    public function jsonSerialize(): mixed
    {
        return $this->toObject();
    }

    /**
     * convert internal state to php array
     *
     * @return array<string,mixed>
     */
    public function toObject()
    {
        $return = [];
        if (sizeof($this->context) == 1) {
            $return['@context'] = array_values($this->context)[0];
        } elseif (sizeof($this->context) > 1) {
            $c = [];
            foreach (array_values($this->context) as $context) {
                $c[] = $context;
            }
            $return['@context'] = $c;
        }
        if (sizeof($this->contexts) > 0) {
            if (array_key_exists('@context', $return)) {
                $return['@context'] =  [$return['@context']];
            } else {
                $return['@context'] =  [];
            }
            $return['@context'][] = $this->contexts;
        }
        if ($this->id !== "") {
            $return['id'] = $this->id;
        }
        $return['type'] = $this->type;

        if ($this->content !== "") {
            $return['content'] = $this->content;
        }

        if ($this->duration !== false) {
            $return['duration'] = $this->duration->format("P%yY%mM%dDT%hH%iM%sS");
        }
        if ($this->height != -1) {
            $return['height'] = $this->height;
        }
        if ($this->href !== "") {
            $return['href'] = $this->href;
        }
        if ($this->endTime > 0) {
            $return['endTime'] = gmdate("Y-m-d\TH:i:s\Z", $this->endTime);
        }
        if ($this->width != -1) {
            $return['width'] = $this->width;
        }
        if (sizeof($this->attachment) > 0) {
            $attachment = [];
            foreach ($this->attachment as $a) {
                 $attachment[] = $a->toObject();
            }
            $return['attachment'] = $attachment;
        }
        if ($this->attributedTo !== "") {
            $return['attributedTo'] = $this->attributedTo;
        }
        if ($this->name !== "") {
            $return['name'] = $this->name;
        }
        if (sizeof($this->icon) > 0) {
            if (sizeof($this->icon) > 1) {
                $icons = [];
                foreach ($this->icon as $icon) {
                    $icons[] = $icon->toObject();
                }
                $return['icon'] = $icons;
            } else {
                $return['icon'] = $this->icon[0]->toObject();
            }
        }
        if (sizeof($this->image) > 0) {
            $images = [];
            foreach ($this->image as $image) {
                $images[] = $image->toObject();
            }
            $return['image'] = $images;
        }
        if ($this->inReplyTo !== "") {
            $return['inReplyTo'] = $this->inReplyTo;
        }
        if ($this->published > 0) {
            $return['published'] = gmdate("Y-m-d\TH:i:s\Z", $this->published);
        }
        if ($this->summary !== "") {
            $return['summary'] = $this->summary;
        }
        if (sizeof($this->tag) > 0) {
            $tags = [];
            foreach ($this->tag as $tag) {
                $tags[] = $tag->toObject();
            }
            $return['tag'] = $tags;
        }
        if ($this->updated > 0) {
            $return['updated'] = gmdate("Y-m-d\TH:i:S\Z", $this->updated);
        }
        if ($this->url !== '') {
            $return['url'] = $this->url;
        }
        if (sizeof($this->to) > 0) {
            $return['to'] = $this->to;
        }
        if (sizeof($this->cc) > 0) {
            $return['cc'] = $this->cc;
        }
        if ($this->mediaType !== '') {
            $return['mediaType'] = $this->mediaType;
        }
        if ($this->object !== null) {
            $return['object'] = $this->object->toObject();
        }
        if ($this->atomURI !== '') {
            $return['atomUri'] = $this->atomURI;
        }
        if ($this->replyAtomURI !== null) {
            $return['inReplyTo'] = $this->replyAtomURI;
            $return['inReplyToAtomUri'] = $this->replyAtomURI;
        }
        if ($this->sensitive !== null) {
            $return['sensitive'] = $this->sensitive;
        }
        if ($this->blurhash !== '') {
            $return['blurhash'] = $this->blurhash;
        }
        if ($this->conversation !== '') {
            $return['conversation'] = $this->conversation;
        }
        if ($this->uuid !== '') {
            $return['uuid'] = $this->uuid;
        }
        return $return;
    }

    /**
     * set uuid
     *
     * @param string $id
     * @return void
     */
    public function setUUID($id)
    {
        $this->uuid = $id;
    }

    public function getUUID() : string
    {
        return $this->uuid;
    }

    public static function parseDateTime(string $input) : int
    {
        $timestamp = 0;
        if (strpos($input, "T")!== false) {
            $date = \DateTime::createFromFormat('Y-m-d\TH:i:sT', $input);
            if ($date === false) {
                $date = \DateTime::createFromFormat('Y-m-d\TH:i:s+', $input);
            }
            if ($date !== false) {
                $timestamp = $date->getTimestamp();
            } else {
                error_log("date parsing error ". $input);
            }
        } else {
            $timestamp = intval($input, 10);
        }
        return $timestamp;
    }
}
